@@ -1,8 +1,10 @@
 import fnmatch
+import functools
 import io
 import ntpath
 import os
 import posixpath
+import re
 import sys
 import weakref
 try:
@@ -11,8 +13,8 @@ except ImportError:
     import dummy_threading as threading
 
 from collections import Sequence, defaultdict
+from contextlib import contextmanager
 from errno import EINVAL, ENOENT, EEXIST
-from functools import wraps
 from itertools import chain, count
 from operator import attrgetter
 from stat import S_ISDIR, S_ISLNK, S_ISREG
@@ -24,7 +26,7 @@ try:
 except ImportError:
     nt = None
 else:
-    if sys.getwindowsversion()[:2] >= (6, 0):
+    if sys.getwindowsversion()[:2] >= (6, 0) and sys.version_info >= (3, 2):
         from nt import _getfinalpathname
     else:
         supports_symlinks = False
@@ -104,9 +106,23 @@ class _NTFlavour(_Flavour):
         {'LPT%d' % i for i in range(1, 10)}
         )
 
+    # Interesting findings about extended paths:
+    # - '\\?\c:\a', '//?/c:\a' and '//?/c:/a' are all supported
+    #   but '\\?\c:/a' is not
+    # - extended paths are always absolute; "relative" extended paths will
+    #   fail.
+
     def splitroot(self, part, sep=sep):
         first = part[0:1]
         second = part[1:2]
+        if (second == sep and first == sep):
+            # XXX extended paths should also disable the collapsing of "."
+            # components (according to MSDN docs).
+            prefix, part = self._split_extended_path(part)
+            first = part[0:1]
+            second = part[1:2]
+        else:
+            prefix = ''
         third = part[2:3]
         if (second == sep and first == sep and third != sep):
             # is a UNC path:
@@ -121,7 +137,10 @@ class _NTFlavour(_Flavour):
                 if index2 != index + 1:
                     if index2 == -1:
                         index2 = len(part)
-                    return part[:index2], sep, part[index2+1:]
+                    if prefix:
+                        return prefix + part[1:index2], sep, part[index2+1:]
+                    else:
+                        return part[:index2], sep, part[index2+1:]
         drv = root = ''
         if second == ':' and first in self.drive_letters:
             drv = part[:2]
@@ -130,7 +149,7 @@ class _NTFlavour(_Flavour):
         if first == sep:
             root = first
             part = part.lstrip(sep)
-        return drv, root, part
+        return prefix + drv, root, part
 
     def casefold(self, s):
         return s.lower()
@@ -147,13 +166,19 @@ class _NTFlavour(_Flavour):
         # Means fallback on absolute
         return None
 
-    def _ext_to_normal(self, s):
-        # Turn back an extended path into a normal DOS-like path
-        if s.startswith(self.ext_namespace_prefix):
+    def _split_extended_path(self, s, ext_prefix=ext_namespace_prefix):
+        prefix = ''
+        if s.startswith(ext_prefix):
+            prefix = s[:4]
             s = s[4:]
             if s.startswith('UNC\\'):
+                prefix += s[:3]
                 s = '\\' + s[3:]
-        return s
+        return prefix, s
+
+    def _ext_to_normal(self, s):
+        # Turn back an extended path into a normal DOS-like path
+        return self._split_extended_path(s)[1]
 
     def is_reserved(self, parts):
         # NOTE: the rules for reserved names seem somewhat complicated
@@ -168,7 +193,7 @@ class _NTFlavour(_Flavour):
         return parts[-1].partition('.')[0].upper() in self.reserved_names
 
 
-_NO_FD = getattr(os, "AT_FDCWD", -100)
+_NO_FD = None
 
 class _PosixFlavour(_Flavour):
     sep = '/'
@@ -219,12 +244,12 @@ class _PosixFlavour(_Flavour):
                     # work to do than the last time.
                     raise ValueError("Symlink loop from %r" % cur)
                 try:
-                    target = accessor.readlinkat(resolved_fd, cur, part)
+                    target = accessor.readlink(cur, part, dir_fd=resolved_fd)
                 except OSError as e:
                     if e.errno != EINVAL:
                         raise
                     # Not a symlink
-                    resolved_fd = accessor.walk_down(resolved_fd, cur, part)
+                    resolved_fd = accessor.walk_down(cur, part, dir_fd=resolved_fd)
                     resolved = cur
                 else:
                     # Take note of remaining work from this symlink
@@ -269,18 +294,16 @@ class _Accessor:
     accessing paths on the filesystem."""
 
 
-# We need all of these
-_at_functions = [
-    'fchmodat', 'fdlistdir', 'fstatat', 'mkdirat', 'openat', 'readlinkat',
-    'renameat', 'symlinkat', 'unlinkat',
-    ]
-supports_openat = all(hasattr(os, fn) for fn in _at_functions)
+if sys.version_info >= (3, 3):
+    supports_openat = (os.listdir in os.supports_fd and
+        os.supports_dir_fd >= { os.chmod, os.stat, os.mkdir, os.open,
+                                os.readlink, os.rename, os.symlink, os.unlink, }
+    )
+else:
+    supports_openat = False
 
 if supports_openat:
 
-    _AT_SYMLINK_NOFOLLOW = os.AT_SYMLINK_NOFOLLOW
-    _AT_REMOVE_DIR = os.AT_REMOVEDIR
-
     def _fdnamepair(pathobj):
         """Get a (parent fd, name) pair from a path object, suitable for use
         with the various *at functions."""
@@ -294,15 +317,15 @@ if supports_openat:
     class _OpenatAccessor(_Accessor):
 
         def _wrap_atfunc(atfunc):
-            @wraps(atfunc)
-            def wrapped(pathobj, *args):
+            @functools.wraps(atfunc)
+            def wrapped(pathobj, *args, **kwargs):
                 parent_fd, name = _fdnamepair(pathobj)
-                return atfunc(parent_fd, name, *args)
+                return atfunc(name, *args, dir_fd=parent_fd, **kwargs)
             return staticmethod(wrapped)
 
         def _wrap_binary_atfunc(atfunc):
-            @wraps(atfunc)
-            def wrapped(pathobjA, pathobjB, *args):
+            @functools.wraps(atfunc)
+            def wrapped(pathobjA, pathobjB, *args, **kwargs):
                 parent_fd_A, nameA = _fdnamepair(pathobjA)
                 # We allow pathobjB to be a plain str, for convenience
                 if isinstance(pathobjB, Path):
@@ -310,37 +333,37 @@ if supports_openat:
                 else:
                     # If it's a str then at best it's cwd-relative
                     parent_fd_B, nameB = _NO_FD, str(pathobjB)
-                return atfunc(parent_fd_A, nameA, parent_fd_B, nameB, *args)
+                return atfunc(nameA, nameB, *args,
+                              src_dir_fd=parent_fd_A, dst_dir_fd=parent_fd_B,
+                              **kwargs)
             return staticmethod(wrapped)
 
-        stat = _wrap_atfunc(os.fstatat)
+        stat = _wrap_atfunc(os.stat)
 
-        def lstat(self, pathobj):
-            return self.stat(pathobj, _AT_SYMLINK_NOFOLLOW)
+        lstat = _wrap_atfunc(os.lstat)
 
-        open = _wrap_atfunc(os.openat)
+        open = _wrap_atfunc(os.open)
 
-        chmod = _wrap_atfunc(os.fchmodat)
+        chmod = _wrap_atfunc(os.chmod)
 
         def lchmod(self, pathobj, mode):
-            return self.chmod(pathobj, mode, _AT_SYMLINK_NOFOLLOW)
+            return self.chmod(pathobj, mode, follow_symlinks=False)
 
-        mkdir = _wrap_atfunc(os.mkdirat)
+        mkdir = _wrap_atfunc(os.mkdir)
 
-        unlink = _wrap_atfunc(os.unlinkat)
+        unlink = _wrap_atfunc(os.unlink)
 
-        def rmdir(self, pathobj):
-            self.unlink(pathobj, _AT_REMOVE_DIR)
+        rmdir = _wrap_atfunc(os.rmdir)
 
         def listdir(self, pathobj):
             fd = self._make_fd(pathobj, tolerant=False)
-            return os.fdlistdir(fd)
+            return os.listdir(fd)
 
-        rename = _wrap_binary_atfunc(os.renameat)
+        rename = _wrap_binary_atfunc(os.rename)
 
-        def symlink(self, target, pathobj):
+        def symlink(self, target, pathobj, target_is_directory):
             parent_fd, name = _fdnamepair(pathobj)
-            os.symlinkat(str(target), parent_fd, name)
+            os.symlink(str(target), name, dir_fd=parent_fd)
 
         def _make_fd(self, pathobj, tolerant=True):
             fd = pathobj._cached_fd
@@ -390,7 +413,7 @@ if supports_openat:
                 return newpath
             parent_fd = pathfd
             for part in parts[:-1]:
-                fd = os.openat(parent_fd, part, os.O_RDONLY)
+                fd = os.open(part, os.O_RDONLY, dir_fd=parent_fd)
                 if parent_fd != pathfd:
                     # Forget intermediate fds
                     os.close(parent_fd)
@@ -403,18 +426,18 @@ if supports_openat:
             return newpath
 
         # Helpers for resolve()
-        def walk_down(self, dirfd, path, name):
-            if dirfd != _NO_FD:
+        def walk_down(self, path, name, dir_fd):
+            if dir_fd != _NO_FD:
                 try:
-                    return os.openat(dirfd, name, os.O_RDONLY)
+                    return os.open(name, os.O_RDONLY, dir_fd=dir_fd)
                 finally:
-                    os.close(dirfd)
+                    os.close(dir_fd)
             else:
                 return os.open(path, os.O_RDONLY)
 
-        def readlinkat(self, dirfd, path, name):
-            if dirfd != _NO_FD:
-                return os.readlinkat(dirfd, name)
+        def readlink(self, path, name, dir_fd):
+            if dir_fd != _NO_FD:
+                return os.readlink(name, dir_fd=dir_fd)
             else:
                 return os.readlink(path)
 
@@ -424,13 +447,13 @@ if supports_openat:
 class _NormalAccessor(_Accessor):
 
     def _wrap_strfunc(strfunc):
-        @wraps(strfunc)
+        @functools.wraps(strfunc)
         def wrapped(pathobj, *args):
             return strfunc(str(pathobj), *args)
         return staticmethod(wrapped)
 
     def _wrap_binary_strfunc(strfunc):
-        @wraps(strfunc)
+        @functools.wraps(strfunc)
         def wrapped(pathobjA, pathobjB, *args):
             return strfunc(str(pathobjA), str(pathobjB), *args)
         return staticmethod(wrapped)
@@ -459,7 +482,17 @@ class _NormalAccessor(_Accessor):
 
     rename = _wrap_binary_strfunc(os.rename)
 
-    symlink = _wrap_binary_strfunc(os.symlink)
+    if nt:
+        if supports_symlinks:
+            symlink = _wrap_binary_strfunc(os.symlink)
+        else:
+            def symlink(a, b, target_is_directory):
+                raise NotImplementedError("symlink() not available on this system")
+    else:
+        # Under POSIX, os.symlink() takes two args
+        @staticmethod
+        def symlink(a, b, target_is_directory):
+            return os.symlink(str(a), str(b))
 
     def init_path(self, pathobj):
         pass
@@ -468,17 +501,145 @@ class _NormalAccessor(_Accessor):
         return None
 
     # Helpers for resolve()
-    def walk_down(self, dirfd, path, name):
-        assert dirfd == _NO_FD
+    def walk_down(self, path, name, dir_fd):
+        assert dir_fd == _NO_FD
         return _NO_FD
 
-    def readlinkat(self, dirfd, path, name):
-        assert dirfd == _NO_FD
+    def readlink(self, path, name, dir_fd):
+        assert dir_fd == _NO_FD
         return os.readlink(path)
 
 _normal_accessor = _NormalAccessor()
 
 
+#
+# Globbing helpers
+#
+
+@contextmanager
+def _cached(func):
+    try:
+        func.__cached__
+        yield func
+    except AttributeError:
+        cache = {}
+        def wrapper(*args):
+            if args in cache:
+                return cache[args]
+            value = cache[args] = func(*args)
+            return value
+        wrapper.__cached__ = True
+        try:
+            yield wrapper
+        finally:
+            cache.clear()
+
+def _make_selector(pattern_parts):
+    pat = pattern_parts[0]
+    child_parts = pattern_parts[1:]
+    if pat == '**':
+        cls = _RecursiveWildcardSelector
+    elif '**' in pat:
+        raise ValueError("Invalid pattern: '**' can only be an entire path component")
+    elif _is_wildcard_pattern(pat):
+        cls = _WildcardSelector
+    else:
+        cls = _PreciseSelector
+    return cls(pat, child_parts)
+
+if hasattr(functools, "lru_cache"):
+    _make_selector = functools.lru_cache()(_make_selector)
+
+
+class _Selector:
+    """A selector matches a specific glob pattern part against the children
+    of a given path."""
+
+    def __init__(self, child_parts):
+        self.child_parts = child_parts
+        if child_parts:
+            self.successor = _make_selector(child_parts)
+        else:
+            self.successor = _TerminatingSelector()
+
+    def select_from(self, parent_path):
+        """Iterate over all child paths of `parent_path` matched by this
+        selector.  This can contain parent_path itself."""
+        path_cls = type(parent_path)
+        is_dir = path_cls.is_dir
+        exists = path_cls.exists
+        listdir = parent_path._accessor.listdir
+        return self._select_from(parent_path, is_dir, exists, listdir)
+
+
+class _TerminatingSelector:
+
+    def _select_from(self, parent_path, is_dir, exists, listdir):
+        yield parent_path
+
+
+class _PreciseSelector(_Selector):
+
+    def __init__(self, name, child_parts):
+        self.name = name
+        _Selector.__init__(self, child_parts)
+
+    def _select_from(self, parent_path, is_dir, exists, listdir):
+        if not is_dir(parent_path):
+            return
+        path = parent_path._make_child_relpath(self.name)
+        if exists(path):
+            for p in self.successor._select_from(path, is_dir, exists, listdir):
+                yield p
+
+
+class _WildcardSelector(_Selector):
+
+    def __init__(self, pat, child_parts):
+        self.pat = re.compile(fnmatch.translate(pat))
+        _Selector.__init__(self, child_parts)
+
+    def _select_from(self, parent_path, is_dir, exists, listdir):
+        if not is_dir(parent_path):
+            return
+        cf = parent_path._flavour.casefold
+        for name in listdir(parent_path):
+            casefolded = cf(name)
+            if self.pat.match(casefolded):
+                path = parent_path._make_child_relpath(name)
+                for p in self.successor._select_from(path, is_dir, exists, listdir):
+                    yield p
+
+
+class _RecursiveWildcardSelector(_Selector):
+
+    def __init__(self, pat, child_parts):
+        _Selector.__init__(self, child_parts)
+
+    def _iterate_directories(self, parent_path, is_dir, listdir):
+        yield parent_path
+        for name in listdir(parent_path):
+            path = parent_path._make_child_relpath(name)
+            if is_dir(path):
+                for p in self._iterate_directories(path, is_dir, listdir):
+                    yield p
+
+    def _select_from(self, parent_path, is_dir, exists, listdir):
+        if not is_dir(parent_path):
+            return
+        with _cached(listdir) as listdir:
+            yielded = set()
+            try:
+                successor_select = self.successor._select_from
+                for starting_point in self._iterate_directories(parent_path, is_dir, listdir):
+                    for p in successor_select(starting_point, is_dir, exists, listdir):
+                        if p not in yielded:
+                            yield p
+                            yielded.add(p)
+            finally:
+                yielded.clear()
+
+
 #
 # Public API
 #
@@ -948,39 +1109,6 @@ class Path(PurePath):
         # A stub for the opener argument to built-in open()
         return self._accessor.open(self, flags, mode)
 
-    def _select_children(self, pattern_parts, recursive):
-        # Helper for globbing
-        # XXX symlink loops
-        if not pattern_parts:
-            yield self
-            return
-        if not self.is_dir():
-            return
-        pat = pattern_parts[0]
-        child_parts = pattern_parts[1:]
-        if _is_wildcard_pattern(pat):
-            cf = self._flavour.casefold
-            for name in self._accessor.listdir(self):
-                name = cf(name)
-                if fnmatch.fnmatchcase(name, pat):
-                    child_path = self._make_child_relpath(name)
-                    for p in child_path._select_children(child_parts, False):
-                        yield p
-                elif recursive:
-                    child_path = self._make_child_relpath(name)
-                    for p in child_path._select_children(pattern_parts, recursive):
-                        yield p
-        else:
-            child_path = self._make_child_relpath(pat)
-            if child_path.exists():
-                for p in child_path._select_children(child_parts, False):
-                    yield p
-            if recursive:
-                for name in self._accessor.listdir(self):
-                    child_path = self._make_child_relpath(name)
-                    for p in child_path._select_children(pattern_parts, recursive):
-                        yield p
-
     # Public API
 
     @classmethod
@@ -1017,7 +1145,8 @@ class Path(PurePath):
         drv, root, pattern_parts = self._flavour.parse_parts((pattern,))
         if drv or root:
             raise NotImplementedError("Non-relative patterns are unsupported")
-        for p in self._select_children(pattern_parts, recursive=False):
+        selector = _make_selector(tuple(pattern_parts))
+        for p in selector.select_from(self):
             yield p
 
     def rglob(self, pattern):
@@ -1028,7 +1157,8 @@ class Path(PurePath):
         drv, root, pattern_parts = self._flavour.parse_parts((pattern,))
         if drv or root:
             raise NotImplementedError("Non-relative patterns are unsupported")
-        for p in self._select_children(pattern_parts, recursive=True):
+        selector = _make_selector(("**",) + tuple(pattern_parts))
+        for p in selector.select_from(self):
             yield p
 
     def absolute(self):
@@ -1186,15 +1316,14 @@ class Path(PurePath):
             self._raise_closed()
         self._accessor.rename(self, target)
 
-    def symlink_to(self, target):
+    def symlink_to(self, target, target_is_directory=False):
         """
         Make this path a symlink pointing to the given path.
         Note the order of arguments (self, target) is the reverse of os.symlink's.
         """
         if self._closed:
             self._raise_closed()
-        # XXX how about target_is_directory?
-        self._accessor.symlink(target, self)
+        self._accessor.symlink(target, self, target_is_directory)
 
     # Convenience functions for querying the stat results
 
