@@ -21,7 +21,7 @@ class _BaseFlavourTest(unittest.TestCase):
     def _check_parse_parts(self, arg, expected):
         f = self.flavour.parse_parts
         sep = self.flavour.sep
-        altsep = self.flavour.sep
+        altsep = self.flavour.altsep
         actual = f([x.replace('/', sep) for x in arg])
         self.assertEqual(actual, expected)
         if altsep:
@@ -111,6 +111,12 @@ class NTFlavourTest(_BaseFlavourTest):
         check(['a', 'Z:\\\\b\\\\c\\', 'd\\'], ('Z:', '\\', ['Z:\\', 'b', 'c', 'd']))
         # UNC paths
         check(['a', '\\\\b\\c\\\\', 'd'], ('\\\\b\\c', '\\', ['\\\\b\\c\\', 'd']))
+        # Extended paths
+        check(['\\\\?\\c:\\'],          ('\\\\?\\c:', '\\', ['\\\\?\\c:\\']))
+        check(['\\\\?\\c:\\a'],         ('\\\\?\\c:', '\\', ['\\\\?\\c:\\', 'a']))
+        # Extended UNC paths (format is "\\?\UNC\server\share")
+        check(['\\\\?\\UNC\\b\\c'],     ('\\\\?\\UNC\\b\\c', '\\', ['\\\\?\\UNC\\b\\c\\']))
+        check(['\\\\?\\UNC\\b\\c\\d'],  ('\\\\?\\UNC\\b\\c', '\\', ['\\\\?\\UNC\\b\\c\\', 'd']))
 
     def test_splitroot(self):
         f = self.flavour.splitroot
@@ -1176,6 +1182,15 @@ class _BasePathTest(unittest.TestCase):
         link.symlink_to(str(target))
         self.assertEqual(link.stat(), target.stat())
         self.assertNotEqual(link.lstat(), target.stat())
+        self.assertFalse(link.is_dir())
+        # Symlinking to a directory
+        target = P['dirB']
+        link = P['dirA', 'linkAAAA']
+        link.symlink_to(target, target_is_directory=True)
+        self.assertEqual(link.stat(), target.stat())
+        self.assertNotEqual(link.lstat(), target.stat())
+        self.assertTrue(link.is_dir())
+        self.assertTrue(list(link))
 
     def test_is_dir(self):
         P = self.cls(BASE)
@@ -1266,13 +1281,13 @@ if pathlib.supports_openat:
             self._readlinkat_fds = []
             self._walk_fds = []
 
-        def readlinkat(self, dirfd, path, name):
-            self._readlinkat_fds.append((dirfd, name))
-            return super().readlinkat(dirfd, path, name)
+        def readlink(self, path, name, dir_fd):
+            self._readlinkat_fds.append((dir_fd, name))
+            return super().readlink(path, name, dir_fd=dir_fd)
 
-        def walk_down(self, dirfd, path, name):
-            self._walk_fds.append((dirfd, name))
-            return super().walk_down(dirfd, path, name)
+        def walk_down(self, path, name, dir_fd):
+            self._walk_fds.append((dir_fd, name))
+            return super().walk_down(path, name, dir_fd=dir_fd)
 
 
 class Mock:
@@ -1289,8 +1304,10 @@ class Mock:
     def __enter__(self):
         def wrapper(*args, **kwargs):
             self.calls += 1
+            self.call_params.append((args, kwargs))
             return self.orig_func(*args, **kwargs)
         self.calls = 0
+        self.call_params = []
         setattr(self.module, self.qualname, wrapper)
         return self
 
@@ -1357,11 +1374,14 @@ class PosixOpenatPathTest(PosixPathTest):
         self.assertEqual(len(pathlib.Path._wrs), n)
 
     def test_iter(self):
-        with Mock("os.fdlistdir") as mock_fdlistdir, \
-             Mock("os.listdir") as mock_listdir:
+        with Mock("os.listdir") as mock_listdir:
             super().test_iter()
-        self.assertEqual(mock_fdlistdir.calls, 1)
-        self.assertEqual(mock_listdir.calls, 0)
+        self.assertEqual(mock_listdir.calls, 1)
+        # A file descriptor was passed
+        args = mock_listdir.call_params[0][0]
+        fd = args[0]
+        self.assertIsInstance(fd, int)
+        self.assertGreaterEqual(fd, 0)
 
     def test_cwd(self):
         p = pathlib.PosixPath.cwd(use_openat=True)
